# -*- coding: utf-8 -*-
"""
Created on Tue Sep 27 17:26:16 2022

@author: beaufils
"""

import os
import numpy as np

path = os.path.join("..","data")

def format_table(year,table,source_name="raw",version_name="formatted",
                 parts = ["FD","T"],
                group_fd=True,ignore_stocks=False,
                fix_empty_rows=True, fill_value = 1,
                sort_countries=False,
                format_output='npy',
                regroup = True):
    """
    Format a table into an MRIO usable by the MRIO class.
    
    The MRIO is loaded from the source_name folder and split into parts.
    By default, final demand is grouped by country.
    Fill_value is used to replace sectors with null gross output, in order
    to avoid numerical errors when computing the Leontief inverse.
    Countries are sorted in alphabetical order using their ISO A3 code.
    Formatted tables are saved in binary files.
    
    Countries and sectors label files will be created automatically if they
    do not exist in the destination folder.

    Parameters
    ----------
    year : int
        Year to load.
    table : str
        Name of the table to load. 
        Currently implemented:
            eora26
            icio2021
    source_name : str, optional
        Name of the folder with the raw data. The default is "raw".
    version_name : str, optional
        Name of the destination folder. The default is "formatted".
    parts : list of str, optional
        Parts of the MRIO to format:
            "FD" for final demand
            "T" for the inter-industry matrix
            "VA" for value added
            "QT" for the satellite extension of intermediate sectors
            "QY" for the satellite extension of final demand
        The default is ["FD","T"].
    group_fd : bool, optional
        Whether to aggregate all elements of final demand for each country. 
        The default is True.
        False is not implemented yet.
    ignore_stocks : bool, optional
        Whether to ignore variations in stocks and inventories and acquisitions
        less disposals of valuables. 
        >Ignoring these categories ensures that final demand is positive.
        If final demand is not to be aggregated (group_fd = False), 
        this parameter is ignored.
        The default is False.
    fix_empty_rows : bool, optional
        Whether to give a value to empty sectors in the raw table. 
        If True, the fill_value is added in the diagonal of the T matrix and 
        of the final demand matrix.
        The default is True.
    fill_value : float-like, optional
        Value to use if fix_emtpy_rows is True. The default is 1.
    sort_countries : bool, optional
        Whether to sort countries by alphabetical order (based on ISO A3 code).
        The default is True.
    format_output : str, optional
        Format under which the formatted tables are to be saved:
            'npy' (default) :  binary format associated with numpy.
            'csv' : human-interpretable but slower
    regroup : bool
        Specific for ICIO.
        If true, subtables for China and Brazil are regrouped.
        In the raw version, China and Brazil have both a table for their 
        domestic supply and for their international production.

    Returns
    -------
    None. Files are saved in the version_name folder in the format specified.

    """
    
    #Build paths
    source = os.path.join(path,table,source_name)
    destination = os.path.join(path,table,version_name)
    
    #Create destination path if needed
    if not os.path.isdir(destination):
        os.mkdir(destination)
    
    #Check wether labels already exist
    if os.path.isfile(os.path.join(destination,"countries.txt")):
        labels=False
        
    else:
        labels = True
    
    #load and pre-format final demand
    tables,countries,sectors = pre_format(table, year, parts, source,
                                          labels, group_fd, ignore_stocks,
                                          regroup)
    
    if labels is False:
        countries = np.loadtxt(os.path.join(destination, "countries.txt"), 
                               dtype=str, delimiter ='\t').tolist()
        sectors = np.loadtxt(os.path.join(destination, "sectors.txt"), 
                               dtype=str, delimiter ='\t').tolist()
    
    t,y = tables["T"],tables["FD"]
    x = np.sum(t,axis=1) + np.sum(y,axis=1)
    if fix_empty_rows:
        log = [[],[]]
        x = np.sum(t,axis=1) + np.sum(y,axis=1)
        for c in range(len(countries)):
            for s in range(len(sectors)):
                if x[c*len(sectors)+s] == 0:
                    log[0].append(countries[c])
                    log[1].append(sectors[s])
                    t[c*len(sectors)+s,c*len(sectors)+s] = fill_value
                    y[c*len(sectors)+s,c] = fill_value
    
    if sort_countries:
        original_order = countries.copy()
        countries.sort()
        for part in parts:
            tables[part] = sort_part(tables[part],part,
                                    original_order,countries,len(sectors))
            
    #Save output table
    for part in tables:
        if format_output == "npy":
            np.save(os.path.join(destination,f"{part}_{year}.{format_output}"),
                    tables[part])
        else:
            np.savetxt(os.path.join(destination,f"{part}_{year}.csv"),
                       delimiter=',')
            
    if labels:
        np.savetxt(os.path.join(destination,"countries.txt"),countries,
                   delimiter='\t',fmt='%s')
        np.savetxt(os.path.join(destination,"sectors.txt"),sectors,
                   delimiter='\t',fmt='%s')

    
def pre_format(table,year,parts,path,labels,group_fd,ignore_stocks,regroup):
    """
    Load and preformat a given table.
    
    The output has the right dimensions.
    Hypothetical numerical procedures are applied in the format_table function.

    Parameters
    ----------
    table : str
        Name of the table to load.
        Currently implemented:
            eora26
    year : str
        Data year to load.
    parts : list of str
        Data blocks to load.
    path : path-like
        Path to raw data.
    labels : bool
        Whether to load label data.
    group_fd : bool
        Whether to group final demand per country.
    ignore_stocks : bool
        Whether to ignore variations in stocks and inventories.
    regroup : bool
        Specific for ICIO.
        If true, subtables for China and Brazil are regrouped.
        In the raw version, China and Brazil have both a table for their 
        domestic supply and for their international production.
    
    Returns
    -------
    tables : dict of 2D numpy array
        Numerical tables.
    countries : list of str
        List of country indices
    sectors : list of str
        List of sectors indices

    """
    if table == "eora26":
        tables, countries, sectors = load_eora26(year,parts,path,labels,
                                                 group_fd, ignore_stocks)
    else:
        print(f"Unkwnown table {table}")
    return tables, countries, sectors
        
def load_eora26(year, parts, path, labels, group_fd, ignore_stocks):
    """
    Load and preformat an EORA26 table.

    Parameters
    ----------
    year : str
        Data year to load.
    parts : list of str
        Data blocks to load.
    path : path-like
        Path to raw data.
    labels : bool
        Whether to load label data.
    group_fd : bool
        Whether to group final demand per country.
    ignore_stocks : bool
        Whether to ignore variations in stocks and inventories.

    Returns
    -------
    tables : dict of {str : 2D numpy array}
        Keys correspond to matrix parts.
        Values are the numerical tables stored as numpy arrays.
    formats : dict of {str : str}
        Format of the table part
    countries : dict of {str : int}
        Dictionnary of country indices
    sectors : dict of {str : int}
        Dictionnary of sectors indices

    """
    
    tables = dict()
    
    for part in parts:
        tables[part] = np.loadtxt(os.path.join(path,
                                             f'Eora26_{year}_bp_{part}.txt'),
                   delimiter = '\t')
        #Exclude statistical discrepancies row and column
        if part == "T":
            tables[part] = tables[part][:189*26,:189*26]
        elif part == "VA":
            tables[part] = tables[part][:,:189*26]
        else:
            if tables[part].ndim == 1:
                tables[part] = np.reshape(tables[part],(1,len(tables[part])))
            tables[part] = tables[part][:189*26,:]
            
        
    if labels:
        labels = np.loadtxt(os.path.join(path, "labels_T.txt"), dtype=str, 
                           delimiter ='\t')
        sectors = labels[:26,3].tolist()
        countries = []
        for i in range(189):
            countries.append(labels[i*26,1][:])
    else:
        countries,sectors = None,None
        
    if "FD" in parts and group_fd:
        fd = tables["FD"]
        fd_new = np.zeros((26*189,189))
        for i in range(189):
            #Ignore last final demand column (statistical discrepancies)
            if ignore_stocks:
                fd_new[:,i] = np.sum(fd[:,i*6:i*6+4],axis=1)
            else:
                fd_new[:,i] = np.sum(fd[:,i*6:(i+1)*6],axis=1)
        tables["FD"] = fd_new[:189*26,:189]
    return tables, countries, sectors

def sort_part(table, part, source, destination, s):
    """
    Reorder the countries of a table.

    Parameters
    ----------
    table : 2D numpy array
        Table to reorder.
    part : str
        Type of part to reorder.
        "T" is treated as a square matrix.
        "FD" is treated with one country per column.
        Other inputs are treated as horizontal matrices.
    source : list of str
        Order of countries in the original table.
    destination : list of str
        Order of countries in the destination table.
    s : int
        Number of sectors.

    Returns
    -------
    new : 2D numpy array
        Newly sorted table.

    """
    new = np.copy(table)
    for old_from,country_from in enumerate(source):
        new_from = destination.index(country_from)
        if part == "T" or part == "FD":
            for old_to,country_to in enumerate(source):
                new_to = destination.index(country_to)
                if part == "T":
                    new[sel(new_from,s), 
                        sel(new_to,s)] = table[sel(old_from,s), sel(old_to,s)]
                elif part == "FD":
                    new[sel(new_from,s),new_to] = table[sel(old_from,s),old_to]
        else:
            new[:,sel(new_from,s)] = table[:,sel(old_from,s)]
    #Checks the consistence of the new table with the original
    assert np.isclose(np.sum(new),np.sum(table))
    return new
            
def sel(index,s):
    """
    Select indices corresponding to a country.

    Parameters
    ----------
    index : int
        Index of the country to select.
    s : int
        Number of sectors.

    Returns
    -------
    slice
        Slice containing the s indices corresponding to the sector.

    """
    return slice(index*s,(index+1)*s)
    
format_table(2016,"eora26")