# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 11:30:04 2022

@author: beaufils
"""

import os
import numpy as np
import pandas as pd

mrio_path = r'C:\Users\beaufils\Documents\Projects\Ressources\MRIO'
zones_path = r'C:\Users\beaufils\Documents\Projects\global_throughflows\data'
crf = '2021_1'
hist = '2.4'

class MRIO:
    """
    Class of an MRIO table.
    
    ...
    
    Attributes
    ------------
    metadata : dict
        Dictionnary storing the metadata: 
            name of the table : str
            version : str
            year : int
            folder : path-like
    c : int
        Number of countries in the table
    s : int
        Number of sectors in the table
    t : 2D numpy array
        Inter-industry matrix
    y : 2D numpy array
        Final demand matrix.
        In the current version, all components of final demand (household cons,
        non-profit serving hh, government cons, capital formation, 
        net stocks and inventories, acquisition less disposal of valuables) 
        are aggregated into one final demand array per country
    qt : 2D numpy array
        Satellite extension for intermediate sectors
    nqt : 2D numpy array
        Normalized satellite extension for intermediate sectors.
    qy : 2D numpy array
        Satellite extension for final demand
    x : 1D numpy array
        Gross output array
    a : 2D numpy array
        Input Coefficient matrix
    l : 2D numpy array
        Leontief inverse
    zones : Dict of {str : list of ints}
        Description of economic zones: 
            Keys are the names of the zones;
            Values are the indices of countries in the corresponding zone.
    zones_lab : list of str
        List of zones' names
    countries : list of str
        List of countries
    sectors : list of str
        List of sectors
    qlab : list of str
        Labels of the environmental extension
    zoned_qt : 3D numpy array
        Environmental extension sorted by zone:
            Axis 0 corresponds of the externality type (cf qlab)
            Axis 1 to the zone
            Axis 2 to the sector-country coordinate
    zoned_nqt : 3D numpy array
        Coefficients of environmental extension sorted by zone:
            Axis 0 corresponds of the externality type (cf qlab)
            Axis 1 to the zone
            Axis 2 to the sector-country coordinate
    zoned_qy : 2D numpy
        Direct externalities aggregated by zone
    zoned_y : 2D numpy array
        Final demand aggregated by zone.
    mrio_path : path-like
        Path to the folder storing MRIO tables
    zones_path : path-like
        Path to the folder storing the zones description.
    zones_file : str
        Name of the file storing the zones description.
        
        
    Methods
    -----------
    load_zones(name) : 
        load new zones
    compute_leontief() :
        Compute the matrix of input coefficients and the leontief inverse
    load_extension(extension):
        Load a new MRIO extension.
    get_TBA_matrix(zone):
        Computes the spatially explicit TBA matrix for a given zone.
    pseudo_diagonlization(part):
        Diagonalize an extension by zone.
    group_y():
        Aggregated final demand by zone.
    extract_zone(part,zone):
        Extract a given zone from a given part of the MRIO table.
    sel(index):
        Returns the slice containing the indices of all sectors of a country.
    """

    def __init__(self,name="icio2021",year=2018,version='formatted',
                 parts=["T","FD"], extension=f"primap_{hist}-{crf}",
                 compute_leontief=True, normalization=True, formatting = True,
                 zones="self", zones_file = "zones_composition.xlsx",
                 zones_path = zones_path,
                 mrio_path=mrio_path,
                 extension_path="same",extension_year="same",
                 extension_labels="labels"):
        """
        Create an MRIO object by loading an MRIO table.
        
        MRIO data should be stored as npy file.

        Parameters
        ----------
        name : str
            Name of the table to load.
        year : int
            Year of the table to load.
        version : str, optional
            Version of the table to load. The default is 'formatted'.
        parts : list of parts to load, optional
            Parts of the table to load:
                "T" for the inter-industry matrix
                "FD" for the final demand matrix
        extension : str, optional
            Name of the environmental extension to load. 
            The default is primap_crf_hist.
        compute_leontief : bool, optional
            Whether to compute and store the Leontief inverse at the loading. 
            The default is True.
        normalization : bool, optional
            Whether to normalized the satellite extension. The default is True.
        formatting : bool, optional
            Whether to create a formatted version of the final demand
            and of the environmental extension. 
            The default is True. 
            Final demand is aggregated by zone, the environmental extension
            is pseudo-diagonalized per zone.
        zones : str, optional
            Composition of economic zones that are treated as single entities. 
            The default is "self", all countries are treated individually.
        zones_file : str, optional
            File where the composition of zones is stored. 
            The default is "zones_composition.xlsx".
        zones_path : path-like, optional
            Path where the zone description file is stored. 
            The default is zones_path.
        mrio_path : path-like, optional
            Path where the MRIO data is stored. The default is mrio_path.
        extension_path : path-like, optional
            Path to the folder where the extension data is stored. 
            The default is same as mrio_path.
        extension_year : int, optional
            Year for which extension data should be loaded. 
            The default is same as MRIO data.
        extension_labels : str, optional
            Name under which the labels of the environmental extension are 
            stored. The default is "labels".

        Returns
        -------
        None.

        """
        folder = os.path.join(mrio_path,name,version)
        assert os.path.isdir(folder), f"Directory {folder} not found."
        
        #Set metadata
        self.metadata =  {
            "table" : name,
            "version" : version,
            "year" : year,
            "folder" : folder,
            }
        
        
        self.zones_path,self.zones_file = zones_path,zones_file
        
        self.countries, self.sectors = [],[]
        
        #Load basic labels
        try:
            self.countries = np.loadtxt(os.path.join(folder,"countries.txt"),
                                        dtype=str, delimiter ='\t').tolist()
            self.sectors = np.loadtxt(os.path.join(folder, "sectors.txt"),
                                      dtype=str, delimiter ='\t')
        except:
            print("Countries and sectors labels not found in {}"\
                  .format(folder))
                
        self.c = len(self.countries)
        self.s = len(self.sectors)
       
        self.load_zones(zones,zones_path,zones_file,False)
        
        #Load basic mrio parts
        table_parts = dict()
        for part in ["T","FD"]:
            file = os.path.join(folder,f"{part}_{year}.npy")
            if part in parts and os.path.isfile(file):
                table_parts[part] = np.load(file)
            else:
                table_parts[part] = None
        self.t,self.y = table_parts.values()
        #TBD: flexible parts importer
        #Requires new part class?
        
        if self.t is not None and self.y is not None:
            x = np.sum(self.t,axis=1) + np.sum(self.y,axis=1)
        else:
            x=None
            normalization = False 
            compute_leontief = False
            #Overrules optional parameters if the gross output is not defined
        self.x = x
        
        #Load extension parts and labels
        if extension is None:
            self.qt,self.nqt = None,None
            self.zoned_qt,self.zoned_nqt = None,None
            self.qy = None
        else:
            self.load_extension(extension,extension_year,extension_path,
                                formatting=formatting,
                                normalization=normalization,
                                labels=extension_labels)
        
        if compute_leontief:
            self.compute_leontief()
        else:
            self.a,self.l,self.nqt = None,None,None
        
    def compute_leontief(self):
        """
        Generate the Leontief inverse for the current MRIO.

        Returns
        -------
        None.

        """
        self.a = np.dot(self.t,np.diag(1/self.x))
        self.l = np.linalg.inv(np.identity(self.c*self.s)-self.a)
                
    def load_zones(self,zones,path=None,file_name=None,formatting=True):
        '''
        Load and treat a file describing economic zones.
        
        Parameters
        ----------
        zones : str
            Name of the zone setting to load.
            If None or "self", countries are taken as zones.
        path : path, optional
            Path to the zones folder.
        file_name : str, optional
            Name of the file storing the zones description
        formatting : bool, optional
            If True, updates the formatting of the final demand (zoned_y) and 
            of the environmental extension (zoned_qt)
        
        
        Returns
        -------
        None.
        Update the following variables:
            zones : dict of {str : list of int}
                Composition of the zones.
                The list of integers correspond to the indices of the countries
                forming the corresponding zone
            zones_lab : list of str
                List of zones names (i.e. keys of the zones dict)

        '''
        if file_name is None:
            file_name = self.zones_file
        if path is None:
            path = self.zones_path
            
        file = pd.read_excel(os.path.join(path,file_name))
        source = file[self.metadata["table"]]
        if zones is None or zones == "self":
            selector = source.copy()
        else:
            selector = file[zones]
        
        mapping = {i:j for i,j in zip(source,selector)}
        
        zones = dict()
        for i,c in enumerate(self.countries):
            if mapping[c] in zones.keys():
                #If zone is already created, add country to the zone
                zones[mapping[c]].append(i)
            else:
                #Otherwise create new zone
                zones[mapping[c]] = [i]
       
        self.zones = zones
        self.zones_lab = list(zones.keys())
        self.zones_lab.sort()
        
        if formatting:
            self.zoned_y = self.group_y()
            self.zoned_qy = self.group_y("qy")
            self.zoned_qt = self.pseudo_diagonalization(self.qt)
            if self.nqt is not None:
                self.zoned_nqt = self.pseudo_diagonalization(self.nqt)
        
    def load_extension(self,name,year="same",folder="same",formatting=True,
                       normalization=True,labels="labels"):
        """
        Load an extension of an MRIO table.
        
        By default, extension is looked into the same folder as the MRIO table,
        year is taken identical, labels are looked for into the "labels" name
        and tables are formatted using the current zones.
        Currently implemented for npy files only.

        Parameters
        ----------
        name : str
            Name of the extension file to load.
        year : int, optional
            Year of environmental data to load. 
            The default is "same", i.e. extension is loaded for the same year
            as the MRIO table.
        folder : path-like, optional
            Path where the environmental extension is stored. 
            The default is "same", i.e. extension is in the folder of the MRIO
            table.
        formatting : boolean, optional
            Whether to do apply the standard formatting to the extension, i.e.
            grouping final demand by zone and diagonalizing the extension 
            by zone.
            The default is True.
        normalizaiton : boolean, optional
            Whether to normalize the extension.
            If true the gross output is used.
        labels : str, optional
            Name of the file storing the labels of the extension. 
            The default is "labels".

        Returns
        -------
        None.

        """
        if year == "same":
            year = self.metadata["year"]
        if folder == "same":
            folder = self.metadata["folder"]
        
        table_parts = dict()
        for part in ["QT","QY"]:
            file = os.path.join(folder,f"{name}_{part}_{year}.npy")
            try:
                table_parts[part] = np.load(file)
            except:
                print(f"{part} part of {name} extension not found.")
                table_parts[part] = None
        self.qt,self.qy = table_parts.values()
        
        qlab = os.path.join(folder,f"{name}_{labels}.txt")
        self.qlab = np.loadtxt(qlab,delimiter='\t',dtype=str,ndmin=1).tolist()
        if normalization:
            self.nqt = self.qt.dot(np.diag(1/self.x))
        else:
            self.nqt = None
        if formatting:
            self.zoned_qy = self.group_y("qy")
            self.zoned_y = self.group_y()
            self.zoned_qt = self.pseudo_diagonalization(self.qt)
            if self.nqt is not None:
                self.zoned_nqt = self.pseudo_diagonalization(self.nqt)
        else:
            self.zoned_y = None
            self.zoned_qt = None
    
    def get_TBA_matrix(self,zone,externality="CO2",normalized=True):
        """
        Computes the explicit throughflow matrix of a given zone.
        
        By default, the normalized version of the satellite extension is used.
        If the satellite extension is natively provided as coefficients,
        the normalized parameter should be set to False.

        Parameters
        ----------
        zone : str
            Name of the zone to compute the throughflow matrix for.
        externality : int or str, optional
            Index of the externality to compute the throughflow for
            or name of the label of the externality to load.
            The default is ["CO2"].
        normalized : bool, optional
            Whether to use the normalized version of the satellite extension
            for the computation. Else, the non-normalized version is used.

        Returns
        -------
        2D numpy matrix
            Spatially explicit throughflow matrix for the chosen country.

        """
        a_ex = self.extract_zone("A",zone)
        y_ex = self.group_y(self.extract_zone("Y",zone))
        if normalized:
            qt_ex = self.pseudo_diagonalization(self.extract_zone("nqt",zone))
        else:
            qt_ex = self.pseudo_diagonalization(self.extract_zone("QT",zone))
        
        if self.zoned_y is None:
            self.zoned_y = self.group_y()
        if self.zoned_qt is None and normalized:
            self.zoned_qt = self.pseudo_diagonalization(self.qt)
        elif self.zoned_nqt is None:
            self.zoned_nqt = self.pseudo_diagonalization(self.nqt)    
        
        if externality in self.qlab:
            if normalized:
                qt = self.zoned_nqt[self.qlab.index(externality)]
            else:
                qt = self.zoned_qt[self.qlab.index(externality)]
            qt_ex = qt_ex[self.qlab.index(externality)]
        elif type(externality) is int:
            if normalized:
                qt = self.zoned_nqt[externality]
            else:
                qt = self.zoned_qt[externality]
            qt_ex = qt_ex[externality]
        else:
            print("Unkown externality specified.")
            print(f"Loaded externalities: {i for i in self.qlab}")
        
        l_ex = np.linalg.inv(np.identity(len(a_ex))-a_ex)
        return qt.dot(self.l).dot(self.zoned_y) - qt_ex.dot(l_ex).dot(y_ex)
        
    def get_throughflow_overview(self,zone,externality="CO2",normalized=True):
        """
        Returns the main characteristics of the throughflow of a given zone.

        By default, the normalized version of the satellite extension is used.
        If the satellite extension is natively provided as coefficients,
        the normalized parameter should be set to False.

        Parameters
        ----------
        zone : str
            Name of the zone to compute the throughflow characteristics for.
        externality : int or str, optional
            Index of the externality to compute the throughflow for
            or name of the label of the externality to load.
            The default is ["CO2"].
        normalized : bool, optional
            Whether to use the normalized version of the satellite extension
            for the computation. Else, the non-normalized version is used.

        Returns
        -------
        pandas DataFrame
            Total of local, exports, imports and traversing flows.
            Includes PBA, CBA and throughflow totals.

        """
        a_ex = self.extract_zone("A",zone)
        y_ex = self.group_y(self.extract_zone("Y",zone))
        if normalized:
            qt_ex = self.pseudo_diagonalization(self.extract_zone("nqt",zone))
        else:
            qt_ex = self.pseudo_diagonalization(self.extract_zone("QT",zone))
        
        if self.zoned_y is None:
            self.zoned_y = self.group_y()
        if self.zoned_qt is None and normalized:
            self.zoned_qt = self.pseudo_diagonalization(self.qt)
        elif self.zoned_nqt is None:
            self.zoned_nqt = self.pseudo_diagonalization(self.nqt)    
        
        if externality in self.qlab:
            if normalized:
                qt = self.zoned_nqt[self.qlab.index(externality)]
            else:
                qt = self.zoned_qt[self.qlab.index(externality)]
            qt_ex = qt_ex[self.qlab.index(externality)]
        elif type(externality) is int:
            if normalized:
                qt = self.zoned_nqt[externality]
            else:
                qt = self.zoned_qt[externality]
            qt_ex = qt_ex[externality]
        else:
            print("Unkown externality specified.")
            print("Loaded externalities:")
            for i in self.qlab:
                print(i)
        
        l_ex = np.linalg.inv(np.identity(len(a_ex))-a_ex)
        l_loc = np.linalg.inv(np.identity(len(self.a))-(self.a-a_ex))
        qt_int = qt - qt_ex
        y_int = self.zoned_y-y_ex
        
        output = pd.DataFrame(index=externality)
        output["Local"] = [np.sum(qt_int.dot(self.l).dot(y_int))]
        output["Direct from final users"] = [self.get_qy(zone,externality)]
        output["Reimports"] = [np.sum(qt_int.dot(self.l-l_loc).dot(y_int))]
        output["Exports"] = [np.sum(qt_int.dot(self.l).dot(y_ex))]
        output["Imports"] = [np.sum(qt_ex.dot(self.l).dot(y_int))]
        output["Traversing"] = [np.sum(qt_ex.dot(self.l-l_ex).dot(y_ex))]
        output["PBA"] = [np.sum(qt_int.dot(self.l).dot(self.zoned_y))]
        output["CBA"] = [np.sum(qt.dot(self.l).dot(y_int))]
        output["TBA"] = [np.sum(qt.dot(self.l).dot(self.zoned_y)\
                               - qt_ex.dot(l_ex).dot(y_ex))]
        return output

    def pseudo_diagonalization(self,part):
        """
        Diagonalize an horizontal exension based on currently defined zones.
        
        Each row of the output holds the externalities associated with a zone.

        Parameters
        ----------
        part : 2D numpy array
            Matrix to pseudo-diagonalize.

        Returns
        -------
        Output : 3D numpy array
            Pseudo diagonalized version of the array.
            Axis:
                0 is the same as the input (i.e. externality type)
                1 is the zone
                2 is the country-sector coordinate

        """
        output = np.zeros((len(part),len(self.zones),len(part[0])))
        for i,zone in enumerate(self.zones_lab):
            for country in self.zones[zone]:
                output[:,i,self.sel(country)] += part[:,self.sel(country)]
        return output
        
    def group_y(self,y='self'):
        """
        Group final demand based on the current zone composition.

        Returns
        -------
        output : 2D numpy array
            Final demand matrix where the final demand from the countries
            forming each zone is grouped into a single column.
        y : 2D numpy array or str
            Final demand to use as raw data.
            If "self", the normal final demand matrix is used.

        """
        if type(y) is str:
            if y == "Y" or y == "y" or y == "self": 
                y = self.y
            elif y == "qy" or y == "QY":
                y = self.qy
            else:
                y=None
        if y is not None:
            output = np.zeros((len(y),len(self.zones)))
            for i,zone in enumerate(self.zones_lab):
                for country in self.zones[zone]:
                    output[:,i] += y[:,country]
            return output
        else:
            return None
    
    def extract_zone(self,part,zone,shape = "square"):
        """
        Extract a zone from a given MRIO part.
        
        Behaves differently based on the shape of the part.
        If "squared" (e.g. an inter-industry matrix), all input and output
        coefficients related to the countries of the zone are set to 0.
        If "horizontal" (e.g. a value added/environmental extension matrix),
        coefficients corresponding to countries of the zone are nullified 
        across all columns.
        If "fd" (e.g. the final demand matrix), only the final demand
        in the countries forming the zone are set to 0.

        Parameters
        ----------
        part : 2D numpy array or str
            Part of the MRIO matrix on which a country should be extracted.
            If the input is a string, the part is looked into the object
            and the shape is guessed:
                "Y" or "FD" : final demand, shape = 'fd'
                "A" : input coefficients, shape = "square"
                "QT" or "Q" : externality matrix, shape = 'horizontal'
                "NQT" or "nqt" : externality coefficients, shape = "horizontal"
                "T" or "Z" : inter-industry matrix, shape ='square'
                "L" : Leontief inverse, shape = 'square'
        zone : str
            Economic zone (countries) to extract.
        shape : str, optional
            Shape of the matrix to be carved. The default is "square".

        Returns
        -------
        output : 2D numpy array
            Part of the matrix without the specified zone.

        """
        if type(part) is str:
            if part == 'FD' or part =='Y':
                part = self.y
                shape = 'fd'
            elif part == "Q" or part == "QT":
                part = self.qt
                shape ='horizontal'
            elif part == "nQ" or part == "nQT" or part == "nqt":
                part = self.nqt
                shape = "horizontal"
            elif part == 'A':
                part = self.a
                shape='square'
            elif part == 'L':
                part = self.l
                shape = 'square'
            elif part == 'T' or part == 'Z':
                part = self.t
                shape = 'square'
            else:
                print(f"Part {part} not found.")
        output = np.copy(part)
        for country in self.zones[zone]:
            if shape == "horizontal" or shape== "square":
                output[:,sel(country,self.s)] = 0
            if shape == "square":
                output[sel(country,self.s),:] = 0
            if shape == "fd":
                output[:,country] = 0
        return output
    
    def get_qy(self, zone, externality = 'CO2'):
        """
        Return the direct externalities caused by final users in a zone.
        
        Input can be a zone name, a country name or an index.

        Parameters
        ----------
        zone : str or int
            Name or index of the country/zone.
        externality : str, optional
            Externality to extract. The default is 'CO2'.

        Returns
        -------
        float
            Volume of externalities caused by the final users in the zone.
            If the zone is uncorrectly specified, no result is returned.

        """
        if externality in self.qlab:
            index = self.qlab.index(externality)
        elif type(externality) is int:
            index = externality
        else:
            print("Unkown externality specified.")
            print(f"Loaded externalities: {i for i in self.qlab}")
        if self.zones_lab is not None and zone in self.zones_lab:
            return self.zoned_qy[index,self.zones_lab.index(zone)]
        elif zone in self.countries:
            return self.qy[index,self.countries.index(zone)]
        elif type(zone) is int:
            return self.zoned_qy[index,zone]
        else:
            print(f"Zone {zone} not found.")
            return None
        
    def get_export_array(self,origin,destination="all",intermediate=True,
                         final=True):
        """
        Return the array of exports from a given country.

        Parameters
        ----------
        origin : str, int or list of ints
            Name of the exporting zone/country or indice(s) of the exporting 
            country.
        destination : str, int or list of ints
            Name of the importing zone/country or indice(s) of the importing 
            country.
            The default is "all".
        intermediate : bool, optional
            Whether to include intermediate trade. The default is True.
        final : bool, optional
            whether to include final trade. The default is True.

        Returns
        -------
        output : numpy array
            Export array.

        """
        output = np.zeros(self.c*self.s)
        if intermediate:
            output += self.get_intermediate_trade(origin, destination,
                                                  shape="exports")
        if final:
            output += self.get_final_trade(origin, destination, 
                                           shape="exports")
        return output
    
    def get_import_array(self,destination,origin="all",intermediate=True,
                         final=True):
        """
        Return the array of imports to a given country/zone.

        Parameters
        ----------
        destination : str, int or list of ints
            Name of the importing zone/country or indice(s) of the importing 
            country.
        origin : str, int or list of ints
            Name of the exporting zone/country or indice(s) of the exporting 
            country.
            The default is "all".
        intermediate : bool, optional
            Whether to include intermediate trade. The default is True.
        final : bool, optional
            whether to include final trade. The default is True.

        Returns
        -------
        output : numpy array
            Import array.

        """
        output = np.zeros(self.c*self.s)
        if intermediate:
            output += self.get_intermediate_trade(origin, destination,
                                                  shape="exports")
        if final:
            output += self.get_final_trade(origin, destination, 
                                           shape="exports")
        return output
    
    def get_intermediate_trade(self, origin, destination, shape="block"):
        """
        Get the directed intermediate trade flows between two countries/zones.
        
        The trade relations can be returned as:
            a block matrix
            a horizontal import array
            a vertical export array
            a scalar total trade volume.
        
        Origin and destinations can be specified based on zones names,
        country names and country indices.
        Intra-national blocks are excluded.
        
        Parameters
        ----------
        origin : str, int or list of ints
            Name of the exporting zone/country or indice(s) of the exporting 
            country.
        destination : str, int or list of ints
            Name of the importing zone/country or indice(s) of the importing 
            country.
        shape : str, optional
            Shape of the desired output:
                "block" : the output has the shape of the inter-industry matrix
                          with only the desired trade block non-null
                "exports" : the output is summed by exporting sector
                "imports" : the output is summed by importing sector
                "total" : the output is the total scalar value
            The default is "block".

        Returns
        -------
        numpy array or float
            Directed intermediate trade flow from the origin to the destination

        """
        #Build list of countries indices for the origin variable
        if self.zones_lab is not None and origin in self.zones_lab:
            origin = self.zones[origin]
        elif origin in self.countries:
            origin = [self.countries.index(origin)]
        elif type(origin) is int:
            origin = [origin]
        elif origin == "all":
            origin = [i for i in range(self.c)]
        elif type(origin) is not list:
            print(f"Exporter zone {origin} not found.")
            return None
        
        #Build list of countries indices for the destination variable
        if self.zones_lab is not None and destination in self.zones_lab:
            destination = self.zones[destination]
        elif destination in self.countries:
            destination = [self.countries.index(destination)]
        elif type(destination) is int:
            destination = [destination]
        elif destination == "all":
            destination = [i for i in range(self.c)]
        elif type(destination) is not list:
            print(f"Exporter zone {destination} not found.")
            return None
        
        #Build an output object of the right shape
        output = np.zeros((self.c*self.s,self.c*self.s))
        for c_origin in origin:
            for c_dest in destination:
                if  c_origin != c_dest:
                    output[self.sel(c_origin),self.sel(c_dest)] +=\
                        self.t[self.sel(c_origin),self.sel(c_dest)]
        if shape == "imports":
            return np.sum(output,axis=0)
        elif shape == "exports":
            return np.sum(output,axis=1)
        elif shape == "total":
            return np.sum(output)
        else:
            return output
        
    def get_final_trade(self, origin, destination, shape="block"):
        """
        Get the directed final trade flows between two countries/zones.
        
        The trade relations can be returned as:
            a block matrix
            a horizontal import array
            a vertical export array
            a scalar total trade volume.
        
        Origin and destinations can be specified based on zones names,
        country names and country indices.
        Intra-national blocks are excluded.
        
        Parameters
        ----------
        origin : str, int or list of ints
            Name of the exporting zone/country or indice(s) of the exporting 
            country.
        destination : str, int or list of ints
            Name of the importing zone/country or indice(s) of the importing 
            country.
        shape : str, optional
            Shape of the desired output:
                "block" : the output has the shape of the final demand matrix
                          with only the desired trade block non-null
                "exports" : the output is an export array
                "imports" : the output in an import array
                "total" : the output is the total scalar value,
            The default is "block".

        Returns
        -------
        numpy array or float
            Directed intermediate trade flow from the origin to the destination

        """
        #Build list of countries indices for the origin variable
        if self.zones_lab is not None and origin in self.zones_lab:
            origin = self.zones[origin]
        elif origin in self.countries:
            origin = [self.countries.index(origin)]
        elif type(origin) is int:
            origin = [origin]
        elif origin == "all":
            origin = [i for i in range(self.c)]
        elif type(origin) is not list:
            print(f"Exporter zone {origin} not found.")
            return None
        
        #Build list of countries indices for the destination variable
        if self.zones_lab is not None and destination in self.zones_lab:
            destination = self.zones[destination]
        elif destination in self.countries:
            destination = [self.countries.index(destination)]
        elif type(destination) is int:
            destination = [destination]
        elif destination == "all":
            destination = [i for i in range(self.c)]
        elif type(destination) is not list:
            print(f"Exporter zone {destination} not found.")
            return None
        
        #Build an output object of the right shape
        output = np.zeros((self.c*self.s,self.c))
        for c_origin in origin:
            for c_dest in destination:
                if  c_origin != c_dest:
                    output[self.sel(c_origin),c_dest] +=\
                        self.y[self.sel(c_origin),c_dest]
        if shape == "imports":
            return np.sum(output,axis=0)
        elif shape == "exports":
            return np.sum(output,axis=1)
        elif shape == "total":
            return np.sum(output)
        else:
            return output
    
    def sel(self,index):
        """
        Select indices corresponding to a country.
        
        Overrides the general sel function

        Parameters
        ----------
        index : int or str
            Index or name of the country to select.

        Returns
        -------
        slice
            Slice containing the s indices corresponding to the country.

        """
        if type(index) is str:
            index = self.countries.where(index)
        return slice(index*self.s,(index+1)*self.s)
    
    def pba(self,zone="all",externality = "CO2",
                                 include_fd=True):
        """
        Returns the Production Based Accounting externalities of an entity.

        Parameters
        ----------
        zone : str or int
            Name or index of the country/zone.
        externality : str, optional
            Externality to extract. The default is 'CO2'.
        include_fd : bool, optional
            Whether to include direct emissions from final users. 
            The default is True.

        Returns
        -------
        total : float
            Volume of externalities.
            If the zone is uncorrectly specified, no result is returned.

        """
        if externality in self.qlab:
            index = self.qlab.index(externality)
        elif type(externality) is int:
            index = externality
        else:
            print("Unkown externality specified.")
            print(f"Loaded externalities: {i for i in self.qlab}")
            return None
        if zone=="all":
            total = np.sum(self.qt[index])
            if include_fd:
                total += np.sum(self.qy[index])
            return total
        else:
            if self.zones_lab is not None and zone in self.zones_lab:
                total = np.sum(self.zoned_qt[index,self.zones_lab.index(zone)])
                if include_fd:
                    total += self.get_qy(zone,externality)
                return total
            elif zone in self.countries:
                total = np.sum(self.qt[index,self.countries.index(zone)])
                if include_fd:
                    total += self.get_qy(zone,externality)
                return total
            elif type(zone) is int:
                total = np.sum(self.zoned_qt[index,zone])
                if include_fd:
                    total += self.get_qy(zone,externality)
                return total
            else:
                print(f"Zone {zone} not found.")
                return None
            
    def cba(self,zone="all",externality = "CO2", include_fd=True, 
            normalized=True):
        """
        Returns the Consumption Based Accounting externalities of an entity.
        
        The normalized parameter should be True except if the original 
        externality extension is already normalized.

        Parameters
        ----------
        zone : str or int
            Name or index of the country/zone.
        externality : str, optional
            Externality to extract. The default is 'CO2'.
        include_fd : bool, optional
            Whether to include direct emissions from final users. 
            The default is True.
        normalized : bool, option
            Whether to use the normalized version of the satellite account.

        Returns
        -------
        total : float
            Volume of externalities.
            If the zone is uncorrectly specified, no result is returned.

        """
        if externality in self.qlab:
            index = self.qlab.index(externality)
        elif type(externality) is int:
            index = externality
        else:
            print("Unkown externality specified.")
            print(f"Loaded externalities: {i for i in self.qlab}")
        if normalized:
            qt = self.nqt
        else:
            qt = self.qt
        if zone=="all":
            total = np.sum(qt[index])
            if include_fd:
                total += np.sum(self.qy[index])
            return total
        
        else:
            if self.zones_lab is not None and zone in self.zones_lab:
                col = self.zones_lab.index(zone)
                total = np.sum(qt[index].dot(self.l).dot(self.zoned_y[:,col]))
                if include_fd:
                    total += self.get_qy(zone,externality)
                return total
            elif zone in self.countries:
                col = self.countries.index(zone)
                np.sum(total = qt[index].dot(self.l).dot(self.y[:,col]))
                if include_fd:
                    total += self.get_qy(zone,externality)
                return total
            elif type(zone) is int:
                total = np.sum(qt[index].dot(self.l).dot(self.zoned_y[:,zone]))
                if include_fd:
                    total += self.get_qy(zone,externality)
                return total
            else:
                print(f"Zone {zone} not found.")
                return None
            
    def spatial_footprint(self,externality = "CO2", include_fd=True, 
            normalized=True):
        """
        Return the spatialize footprint matrix.
        
        This matrix contain the direct and indirect emissions caused in any
        zone (in row) by the final demand in any other zone (in column).

        Parameters
        ----------
        externality : str, optional
            Externality to extract. The default is 'CO2'.
        include_fd : bool, optional
            Whether to include direct emissions from final users. 
            The default is True.
        normalized : bool, option
            Whether to use the normalized version of the satellite account.

        Returns
        -------
        output : 2D numpy matrix
            Spatialized footprint matrix.

        """
        if externality in self.qlab:
            index = self.qlab.index(externality)
        elif type(externality) is int:
            index = externality
        else:
            print("Unkown externality specified.")
            print(f"Loaded externalities: {i for i in self.qlab}")
        if normalized:
            qt = self.zoned_nqt
        else:
            qt = self.zoned_qt
        output = qt[index].dot(self.l).dot(self.zoned_y)
        if include_fd:
            for i in range(len(self.zones_lab)):
                output[i,i] += self.zoned_qy[index,i]
        return output
                
            
    def rename_zone(self,zone,new_name,
                    save_under=None, path=None, file_name=None,
                    formatting = True, load = True):
        """
        Rename a given zone.
        
        Does not change the zone composition.
        An error is returned if the zone neame does not exist.

        Parameters
        ----------
        zone : str
            Original name of the zone to rename.
        new_name : str
            New name for the zone.
        save_under : str, optional
            Name under which the new zone is saved in the current zone file. 
            The default is None: the new zone is not saved.
        path : path-like, optional
            Path to the folder containing the economic zones description. 
            The default is the current zone path.
        file_name : str, optional
            Name of the file containing the zones description. 
            The default is the current zone file.
        formatting : bool, optional
            If True, updates the formatting of the final demand (zoned_y) and 
            of the environmental extension (zoned_qt).
            Formatting is only done if the new zone composition is loaded.
        load : bool, optional
            Whether to update the current zone composition.
            The default is True.

        Returns
        -------
        None.

        """
        new_zones = self.zones.copy()
        if zone in new_zones.keys():
            new_zones[new_name] = self.zones[zone]
            new_zones.pop(zone)
        else:
            print(f"Unkown zone {zone}")
        
        if save_under is not None:
            self.write_zones(new_zones, path=path, file_name=file_name,
                             name=save_under)
        if load:
            self.zones = new_zones
            self.zones_lab = list(new_zones.keys())
            self.zones_lab.sort()
            
            if formatting:
                self.zoned_y = self.group_y()
                self.zoned_qy = self.group_y("qy")
                self.zoned_qt = self.pseudo_diagonalization(self.qt)
                if self.nqt is not None:
                    self.zoned_nqt = self.pseudo_diagonalization(self.nqt)
    
    def modify_zone(self, zone, new_zone, mode="add", 
                    save_under=None, path=None, file_name=None,
                    formatting = True, load = True):
        """
        Add a zone or delete a country to/from a given zone.
        
        If the new zone does not exist, the zone to be added is simply renamed
        as the new zone.
        The new zones composition is then either saved 
        using the write_zones method or simply loaded as the new_zone.

        Parameters
        ----------
        zone : str
            Name of the zone to which a country is added.
        country : str
            Name of the country to add to the zone.
        mode : str, optional
            Whether to add (add) or to delete (del) a country to/from the zone.
        save_under : str, optional
            Name under which the new zones setting is saved in the zone file. 
            The default is None: the new zone is not saved.
        path : path-like, optional
            Path to the folder containing the economic zones description. 
            The default is the current zone path.
        file_name : str, optional
            Name of the file containing the zones description. 
            The default is the current zone file.
        formatting : bool, optional
            If True, updates the formatting of the final demand (zoned_y) and 
            of the environmental extension (zoned_qt).
            Formatting is only done if the new zone composition is loaded.
        load : bool, optional
            Whether to update the current zone composition.
            The default is True.

        Returns
        -------
        None.

        """
        new_zones = self.zones.copy()
        if new_zone in new_zones.keys():
            if mode == "add":
                new_zones[new_zone] += self.zones[zone]
                new_zones.pop(zone)
            else:
                if self.countries.index(zone) in new_zones.new_zone:
                    new_zones[zone] = [self.countries.index(zone)]
                    new_zones[new_zone].remove(self.countries.index(zone))
                else:
                    print(f"Country {zone} not found in zone {new_zone}")
        else:
            if mode == "add":
                new_zones[new_zone] = self.zones[zone]
                new_zones.pop(zone)
            else:
                print(f"Zone {zone} not found.")
                
        if save_under is not None:
            self.write_zones(new_zones, path=path, file_name=file_name,
                             name=save_under)
        if load:
            self.zones = new_zones
            self.zones_lab = list(new_zones.keys())
            self.zones_lab.sort()
            
            if formatting:
                self.zoned_y = self.group_y()
                self.zoned_qy = self.group_y("qy")
                self.zoned_qt = self.pseudo_diagonalization(self.qt)
                if self.nqt is not None:
                    self.zoned_nqt = self.pseudo_diagonalization(self.nqt)
            
        
    def write_zones(self,zones="self",
                    path=None,file_name=None,name="new zone"):
        """
        Create a new zone in a zone description file.
        
        The current zone file is used by default.
        If a new file name is specified, a new country file is created
        with the source file composition and the new zone.

        Parameters
        ----------
        zones : dict of str : list of ints, optional
            Description of the economic zones to save.
            The default is the currently define zone composition.
        path : path-like, optional
            Path to the folder containing the economic zones description. 
            The default is the current zone path.
        file_name : str, optional
            Name of the file containing the zones description. 
            The default is the current zone file.
            If a new name is defined, a new zone file is created and defined
            as the default zone file.
        name : str, optional
            Name of the new zones composition. The default is "new zone".

        Returns
        -------
        None.

        """
        if path is None:
            path = self.zones_path
            
        if zones is None or zones == "self":
            zones = self.zones
            
        file = pd.read_excel(os.path.join(self.zones_path,self.zones_file))
        source = file[self.metadata["table"]].to_numpy()
        new_zone = source.copy()
        for zone in zones.keys():
            for country in zones[zone]:
                new_zone[source==self.countries[country]] = zone
        if file_name is not None:
            self.zones_path = path
            self.zones_file = file_name
            file = pd.DataFrame()
            file[self.metadata["table"]] = source
        file[name] = new_zone
        file.to_excel(os.path.join(self.zones_path,self.zones_file),
                          index=False)
            
        
def sel(index,s):
    """
    Select indices corresponding to a country.

    Parameters
    ----------
    index : int
        Index of the country to select.
    s : int
        Number of sectors.

    Returns
    -------
    slice
        Slice containing the s indices corresponding to the country.

    """
    return slice(index*s,(index+1)*s)